"""Init"""
import sys

from geochron._version import __version__  # noqa: F401
from geochron.utils.conditional_imports import ConditionalPackageInterceptor
from geochron.chronnet import convert_chronnet
from geochron.timehex import convert_timehex
from geochron.time_grid import convert_time_grid
from geochron.geotimehash import convert_geotimehash
from geochron.geosynchnet import convert_geosynchnet

ConditionalPackageInterceptor.permit_packages(
    {
        'networkx': 'networkx>=3.0,<4.0',
        'timehash': 'timehash>=1.2,<2',
        'branca': 'branca>=0.7.2,<1.0',
    }
)
sys.meta_path.append(ConditionalPackageInterceptor)  # type: ignore

__all__ = [
    'convert_chronnet',
    'convert_timehex',
    'convert_geotimehash',
    'convert_geosynchnet',
    'convert_time_grid' 
]




import datetime as dt
from geochron.time_slicing import get_timestamp_intervals, time_slice_track
from geostructures import Coordinate, GeoPoint
from geostructures.collections import  Track

track = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5)),
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
)

def test_get_timestamp_intervals():
    timestamps = get_timestamp_intervals(track,dt.timedelta(hours=1))
    
    assert timestamps[-1] == dt.datetime(2020, 1, 1, 10, 5,1, tzinfo=dt.timezone.utc)



def test_time_slice_track():
    test_timestamps = [dt.datetime(2020, 1, 1, 9, 5, tzinfo=dt.timezone.utc),
    dt.datetime(2020, 1, 1, 10, 5, 1, tzinfo=dt.timezone.utc)]
    track_list = time_slice_track(track, test_timestamps)

    assert track_list[-1].end == dt.datetime(2020, 1, 1, 10, 5, tzinfo=dt.timezone.utc)

    


import pytest
import math
from datetime import datetime, timedelta

from geostructures import Coordinate, GeoPoint
from geostructures.collections import  Track

from geochron.time_grid import  break_time_interval, round_down_datetime, extract_intervals_in_range \
,create_time_list_from_datetimes, convert_time_grid 

@pytest.fixture
def sample_track():
    return Track(
        [
            GeoPoint(Coordinate(-0.104154, 51.511920), dt=datetime(2020, 1, 1, 8, 5)),
            GeoPoint(Coordinate(-0.096533, 51.511903), dt=datetime(2020, 1, 1, 9, 23)),
            GeoPoint(Coordinate(-0.083765, 51.514423), dt=datetime(2020, 1, 1, 9, 44)),
            GeoPoint(Coordinate(-0.087478, 51.508595), dt=datetime(2020, 1, 1, 10, 11)),
        ]
    )
def test_break_time_interval(sample_track):
    time_interval = timedelta(hours=1)
    interval_list = [
        datetime(2020, 1, 1, 8, 0),
        datetime(2020, 1, 1, 9, 0),
        datetime(2020, 1, 1, 10, 0),
    ]
    
    result = break_time_interval(sample_track, interval_list, time_interval)
    
    # Check the length of the result
    assert len(result) == 3
    

    assert result[0] == Track([GeoPoint(Coordinate(-0.104154, 51.511920), dt=datetime(2020, 1, 1, 8, 5))])
    
    # Ensure the end times are handled correctly
    assert result[2] == Track([GeoPoint(Coordinate(-0.087478, 51.508595), dt=datetime(2020, 1, 1, 10, 11))])


def test_round_down_datetime():
    # Test cases
    test_cases = [
        (datetime(2023, 1, 1, 8, 5), timedelta(hours=1), datetime(2023, 1, 1, 8, 0)),
        (datetime(2023, 1, 1, 8, 59), timedelta(hours=1), datetime(2023, 1, 1, 8, 0)),
        (datetime(2023, 1, 1, 8, 30), timedelta(minutes=15), datetime(2023, 1, 1, 8, 30)),
        (datetime(2023, 1, 1, 8, 45), timedelta(minutes=30), datetime(2023, 1, 1, 8, 30)),
        (datetime(2023, 1, 1, 8, 0, tzinfo=None), timedelta(hours=1), datetime(2023, 1, 1, 8, 0)),
        (datetime(2023, 1, 1, 8, 0), timedelta(days=1), datetime(2023, 1, 1, 0, 0)),
    ]
    
    for dt, delta, expected in test_cases:
        assert round_down_datetime(dt, delta) == expected

def test_extract_intervals_in_range():
    # Define test cases
    start_time = datetime(2023, 1, 1, 0, 0, 0)
    end_time = datetime(2023, 1, 1, 6, 0, 0)
    interval = timedelta(hours=2)
    
    expected_intervals = [
        datetime(2023, 1, 1, 0, 0, 0),
        datetime(2023, 1, 1, 2, 0, 0),
        datetime(2023, 1, 1, 4, 0, 0),
        datetime(2023, 1, 1, 6, 0, 0)
    ]

    # Execute the function
    result_intervals = extract_intervals_in_range(start_time, end_time, interval)

    # Check that the results match the expected intervals
    assert result_intervals == expected_intervals


def test_create_time_list_from_datetimes():
    start_datetime = datetime(2023, 1, 1, 0, 0, 0)
    num_intervals = 5
    interval = timedelta(hours=1)
    
    expected_times = [
        datetime(2023, 1, 1, 0, 0, 0),
        datetime(2023, 1, 1, 1, 0, 0),
        datetime(2023, 1, 1, 2, 0, 0),
        datetime(2023, 1, 1, 3, 0, 0),
        datetime(2023, 1, 1, 3, 59, 59)  # Adjusted by subtracting one second
    ]
    
    result = create_time_list_from_datetimes(start_datetime, num_intervals, interval)
    
    assert result == expected_times

# Mock hash function for testing
class MockHashFunc:
    def hash_coordinates(self, points):
        return {f"hash_{i}": point for i, point in enumerate(points)}

def test_convert_time_grid(sample_track):
    time_interval = timedelta(hours=1)
    time_subinterval = timedelta(minutes=30)
    hash_func = MockHashFunc()
    result = convert_time_grid(sample_track, time_interval, time_subinterval, hash_func.hash_coordinates)

    # Expected DataFrame structure
    expected_columns = ['Period_1', 'Period_2', 'interval_start']
    assert list(result.columns) == expected_columns
    assert result['Period_1'][0] == 'hash_0'
    assert result['Period_2'][0] == 0 






"Representation as a time grid"
import math
from datetime import  date, datetime, timedelta
from typing import Callable, List
import pandas as pd
from geostructures.collections import FeatureCollection, Track
from geochron.time_slicing import get_timestamp_intervals, time_slice_track
from geostructures.time import TimeInterval


def break_time_interval(track: Track, interval_list: List, time_interval: timedelta):
    """
    Breaks a Track into several tracks partitioned by a time interval.
    
    Args:
        track: The target geostructures Track
        time_interval: A timedelta representing the interval to partition by

    Returns:
        A list of tracks
    """
    timestamps = interval_list.copy()
    timestamps.append(interval_list[-1] + time_interval)
    track_list = time_slice_track(track, timestamps)
    del track_list[0]
    return track_list



def round_down_datetime(dt:datetime, delta:timedelta):
    """
    Rounds down a datetime object to the nearest interval of the specified timedelta.
    
    Args:
        dt: The datetime object to be rounded down.
        delta: A timedelta object representing the interval to round down to.

    Returns:
        A datetime object rounded down to the nearest interval of the specified timedelta.
    """
    if dt.tzinfo is None:
        min_dt = datetime.min
    else:
        min_dt = datetime.min.replace(tzinfo=dt.tzinfo)

    remainder = dt - (dt - min_dt) % delta
    return remainder

def extract_intervals_in_range(start_time: datetime, end_time: datetime, interval: timedelta):
    """
    Extracts intervals within a specified range by rounding down to the nearest interval.
    
    Args:
        start_time: A datetime object marking the start of the range.
        end_time: A datetime object marking the end of the range.
        interval: A timedelta object representing the interval to round down to.

    Returns:
        A list of datetime objects representing the intervals rounded 
        down to the nearest specified interval within the range.
    """
    current_time = start_time
    intervals_list = []

    while current_time <= end_time:
        intervals_list.append(round_down_datetime(current_time, interval))
        current_time += interval

    # Add an extra interval to cover the end time if necessary
    if intervals_list[-1] < end_time:
        intervals_list.append(round_down_datetime(current_time, interval))

    return intervals_list




def create_time_list_from_datetimes(start_datetime: datetime, num_intervals: int, interval: timedelta):
    """
    Generates a list of datetime objects from a given start time, creating a specified number of intervals.
    
    Args:
        start_datetime: A datetime object marking the start time.
        num_intervals: The number of intervals to create.
        interval: A timedelta object representing the interval between times.

    Returns:
        A list of datetime objects representing each interval.
    """
    times = []
    current_time = start_datetime

    for _ in range(num_intervals):
        times.append(current_time)
        current_time += interval
    times[-1] = times[-1]- timedelta(seconds=1)
    return times




def convert_time_grid(fcol: FeatureCollection,
                      time_interval: timedelta,
                      time_subinterval: timedelta,
                      hash_func: Callable,
                      integerize=False):
    """
    Converts a track into a time grid dataframe,
    partitioning it by specified time intervals and subintervals.
    
    Args:
        track: The target geostructures Track to be converted.
        time_interval: A timedelta object representing the primary interval to partition the track.
        time_subinterval: A timedelta object representing the subinterval.
        hash_func: A callable function used to hash coordinates of track centroids. Must use hash_coordinates
        integerize: A boolean flag indicating whether to convert hashed values to integers.

    Returns:
        A pandas DataFrame representing the time grid with track segments hashed into intervals (rows) and subintervals (columns).
    """
    all_tracks = []
    track = Track(fcol.geoshapes)
    num_intervals = math.ceil(time_interval / time_subinterval)
    columns = [f"Period_{i+1}" for i in range(num_intervals)]
    interval_list = extract_intervals_in_range(track.start, track.end, time_interval)
    track_list = break_time_interval(track,interval_list, time_interval)

    for tr, interval in zip(track_list, interval_list):
        time_list = create_time_list_from_datetimes(interval, num_intervals+1, time_subinterval)
        interval_tracks = []
        for i in range(len(time_list) - 1):
            current_time = time_list[i] - timedelta(microseconds=1)
            next_time = time_list[i + 1] - timedelta(microseconds=1)
            subinterval_track = tr.filter_by_dt(TimeInterval(current_time,next_time))
            if len(subinterval_track) > 0:
                point = subinterval_track.centroid
                hash_value = list(hash_func([point]).keys())[0]
                if integerize:
                    interval_tracks.append(int(hash_value, 16))
                else:
                    interval_tracks.append(hash_value)
            else:
                interval_tracks.append(0)
        all_tracks.append(interval_tracks)

    time_grid = pd.DataFrame(all_tracks, columns=columns)
    time_grid['interval_start'] = interval_list
    return time_grid



"""Folium helpers"""
import json
from typing import Any, Callable, Optional, Union
import h3
import pandas as pd
from branca.colormap import LinearColormap
from geostructures.geohash import h3_to_geopolygon
from geostructures import FeatureCollection

def timehex_styledict(timehex: pd.DataFrame, opacity= float(.7), cmap:Optional[Callable] = None):
    """
    Creates a styledict appropriate for a folium TimeSliderChoropleth from
    a timehex dataframe

    Args:
        timehex: A timehex pandas dataframe 

    Returns:
        A style dictionary suitable to use with folium's 
        TimeSliderChoropleth
    """
    timehex_duplicate = timehex.copy()
    timehex_duplicate['time'] = timehex_duplicate['start_time'].astype("int64") // 10 ** 9

    select_timehex = timehex_duplicate.drop(['interval', 'start_time', 'end_time'], axis=1)

    timehex_long = select_timehex.melt(id_vars='time', var_name='id', value_name='value')

    timehex_long = timehex_long.fillna(0)

    #color scale
    max_color = max(timehex_long['value'])
    min_color = min(timehex_long['value'])
    if cmap is None:
        used_cmap = LinearColormap(colors=['blue','red'], vmin=min_color, vmax=max_color)
    elif isinstance(cmap, list):
        used_cmap = LinearColormap(colors= cmap, \
        vmin=min_color, vmax=max_color)
    else:
        used_cmap = cmap


    # Vectorized operations for creating styledict
    styledict = timehex_long.groupby('id').apply(lambda df: df.set_index('time')['value']\
    .apply(lambda x: {'color': used_cmap(x), 'opacity': 0 if x == 0 else opacity})\
    .to_dict()).to_dict()

    return styledict



def h3_to_geojson(h3_hashmap):
    """
    Converts a H3 hashmap to a GeoJSON FeatureCollection

    Args:
        h3_hashmap: A H3 hashmap

    Returns:
        A GeoJSON FeatureCollection as a JSON string 
    """
    # Initialize an empty list to store the features
    features = []

    # Iterate over the H3 hashmap
    for h3_index, properties in h3_hashmap.items():
        # Convert the H3 index to a GeoJSON polygon
        geojson_polygon = h3.h3_to_geo_boundary(h3_index, geo_json=True)

        # Create a GeoJSON feature for the H3 index
        feature = {
            "type": "Feature",
            "geometry": {
                "type": "Polygon",
                "coordinates": [geojson_polygon]
            },
            "properties": properties,
            "id" : h3_index
        }

        # Add the feature to the list of features
        features.append(feature)

    # Create a GeoJSON FeatureCollection
    feature_collection = {
        "type": "FeatureCollection",
        "features": features
    }

    # Return the FeatureCollection as a JSON string
    return json.dumps(feature_collection)

def timehex_backgroundata(timehex: pd.DataFrame):
    """
    Formats backgroundata polygons appropriate for a folium TimeSliderChoropleth from
    a timehex dataframe

    Args:
        timehex: A timehex dataframe

    Returns:
        a GeoJson FeatureCollection
    """
    select_timehex = timehex.drop(['interval', 'start_time', 'end_time'], axis=1)
    hashmap = select_timehex.to_dict()
    backgroundata = h3_to_geojson(hashmap)

    return backgroundata

def normalize(value: float, min_val: float, max_val: float):
    """
    Normalize a value on a scale from 1 to 0 based on a given
    min and max value.

    Args:
        value: a value
        min_val: the minimum value
        max_val: the maximum value 

    Returns:
        a scaled value between 1 to 0
    """
    if max_val == min_val:
        raise ValueError("max_val and min_val cannot be equal (division by zero)")
    scaled_value = (value - min_val) / (max_val - min_val)
    return scaled_value

def constant_return(value: float, constant: Any):
    """
    Returns a constant when fed a value  
    Args:
        value: any value
        constant: any constant
    Returns:
        the specified constant 
    """
    return constant

def add_hashmap_properties(original_hashmap:dict, time:pd.Timestamp, opacity: Union[float, str] \
                            , cmap: Union[Optional[Callable], list] = None):
    """
    Adds properties to the hashmap necessary for display.

    Args:
        original_hashmap: a hashmap
        time: time of hashmap
        cmap: a Branca colormap or a list of colors 
        opacity: desired opacity of shapes takes key word gradient
        to vary opacity based on weight of shape in the time period
    Returns:
        a GeoJson FeatureCollection
    """
    new_dict = {}
    removed_empty = {key: value for key, value in original_hashmap.items() if value != 0}

    if len(removed_empty) == 0:
        color_min = 0
        color_max = 0
    else:
        max_key = max(removed_empty, key=removed_empty.get) # type: ignore[arg-type]
        min_key = min(removed_empty, key=removed_empty.get) # type: ignore[arg-type]
        color_min = removed_empty[min_key]
        color_max = removed_empty[max_key]

    if cmap is None:
        used_cmap = LinearColormap(colors=['blue','red'], \
        vmin= color_min, vmax= color_max)
    elif isinstance(cmap, list):
        used_cmap = LinearColormap(colors= cmap, \
        vmin= color_min, vmax= color_max)
    else:
        used_cmap = cmap

    if opacity == "gradient":
        used_opacity = lambda x: normalize(x,min_val = color_min, max_val= color_max)
    else:
        used_opacity = lambda x: constant_return(x, constant= opacity)

    for key, value in removed_empty.items():
        new_dict[key] = {'popup': 'weight= ' + str(value) +
                    '<br> center(lat,lon)= ' + str(h3.h3_to_geo(key)), 
                    'time': time,'style':{'opacity': used_opacity(value), 'color': used_cmap(value),
                                               'fillOpacity': used_opacity(value)}}
    return new_dict

def timehex_timestampedgeojson(timehex: pd.DataFrame, opacity: Union[float, str] = .7 \
                               , cmap: Union[Optional[Callable], list] = None):
    """
    Formats a timehex into the correct data format for Folium's timestampedgeojson

    Args:
        timehex: A timehex dataframe
        opacity: desired opacity of shapes takes key word gradient
        to vary opacity based on weight of shape in the time period
        cmap: a Branca colormap or a list of colors 
    Returns:
        a GeoJson FeatureCollection
    """
    start_time_list = timehex['start_time'].tolist()
    select_timehex = timehex.drop(['interval', 'start_time', 'end_time'], axis=1).fillna(0)
    list_hashmaps = select_timehex.to_dict('records')
    polygon_list:list = []

    for hashmap, start_time in zip(list_hashmaps, start_time_list):
        hashmap_properties= add_hashmap_properties(original_hashmap = hashmap,\
        time = start_time, opacity = opacity, cmap= cmap)
        polygons = {h3_to_geopolygon(k, properties= v) for k,v in hashmap_properties.items()}
        polygon_list.extend(polygons)

    return FeatureCollection(polygon_list).to_geojson()



""" Representation as chronnets """
from datetime import  timedelta
import itertools
from typing import Callable, List
import numpy as np
import pandas as pd

from geostructures.collections import  FeatureCollection,Track
from geochron.time_slicing import get_timestamp_intervals, time_slice_track


def hash_tracks_into_netdf(track_list: List, timestamps: List, hash_func: Callable):
    """
    Converts a list of tracks into a pandas dataframe suitable for chron net creation
    using a specified hashing function with intervals reflected
    in a corresponding timestamp list 
    
    Args:
        track_list: a list of tracks broken down by equal intervals

        timestamps: a list of corresponding timestamps

        hash_func: the hashing function

    Returns:
        A pandas dataframe
    """
    interval_start = track_list[0].start
    master_list = []
    for track,timestamp  in zip(track_list, timestamps):
        hashmap = hash_func(track)
        begin_string= interval_start.strftime("%Y-%m-%d %H:%M:%S")
        ending_string= timestamp.strftime("%Y-%m-%d %H:%M:%S")
        interval = begin_string + ", " + ending_string
        hexes = list(hashmap.keys())
        counts = list(hashmap.values())
        hex_list = [element for element, count in zip(hexes, counts) for _ in range(count)]
        tuples_list = [(item, interval) for item in hex_list]
        master_list += tuples_list
        interval_start = timestamp


    df = pd.DataFrame(master_list, columns=['cell', 'time'])

    return df


def chronnet_create(df: pd.DataFrame, self_loops: bool, mode= str):
    """
    Converts a properly formatted pandas dataframe with the columns
    of cell and time to a networkx network where nodes are locations
    and edges are formed between nodes with consecutive times. 
    
    Args:
        df: a pandas dataframe withe two columns cell and time

        self_loops: whether self loops are included in the network

        mode: whether the network is directed or undirected

    Returns:
        A networkx network
    """
    # pylint: disable=import-outside-toplevel
    import networkx as nx # type: ignore
    time_seq = sorted(np.unique(df['time']))
    if len(time_seq) < 2: # pragma: no cover
        print("The total time interval in the dataset should be larger than two.")
    links = pd.DataFrame()
    cells_before=[]
    cells_after=[]
    weights = []
    for i in range(len(time_seq)-1):
        time1 = df[df.time == time_seq[i]]['cell'].tolist()
        time2 = df[df.time == time_seq[i+1]]['cell'].tolist()
        connections = list(itertools.product(time1, time2))
        cells_before.extend([connection[0] for connection in connections])
        cells_after.extend([connection[1] for connection in connections])
        weights.extend([1]*len(connections))
    links['from'] = cells_before
    links['to'] = cells_after
    links['weight'] = weights
    links = links.sort_values(by=['from', 'to']).reset_index().drop(columns = ['index'])
    links = links.groupby(['from', 'to'], as_index=False)['weight'].sum().reset_index()
    net = nx.DiGraph()
    if len(links) !=0:
        net.add_nodes_from(np.unique(df['cell']))
        edgelist = []
        for index, rows in links.iterrows():
            edgelist.append(tuple([rows['from'], rows['to'], rows['weight']]))
        net.add_weighted_edges_from(edgelist)
        if self_loops is False:
            net.remove_edges_from(list(nx.selfloop_edges(net)))
        if mode == 'undirected':
            net = net.to_undirected()
    else: # pragma: no cover
        print("Empty graph returned.")

    return net


def convert_chronnet(fcol: FeatureCollection, time_delta: timedelta,
     hash_func: Callable, self_loops: bool, mode: str):
    """
    Converts a FeatureCollection into a chronnet with a specified time interval
    using a specified hashing function
    
    Args:
        fcol: a FeatureCollection with time bound shapes 

        time_delta: the desired time interval

        hash_func: the hashing function

        self_loops: whether self loops are included in the network

        mode: whether the network is directed or undirected

    Returns:
        A networkx network 
    """
    track = Track(fcol.geoshapes)
    timestamps = get_timestamp_intervals(track, time_delta)

    track_list = time_slice_track(track, timestamps)

    df = hash_tracks_into_netdf(track_list, timestamps, hash_func)

    chronnet = chronnet_create(df, self_loops, mode)

    return chronnet



""" Representation as geosynchnet"""
from datetime import  timedelta
from typing import Callable
import pandas as pd

from geostructures.collections import  FeatureCollection,Track
from geochron.chronnet import hash_tracks_into_netdf
from geochron.time_slicing import get_timestamp_intervals, time_slice_track

def geosynchnet_create(df: pd.DataFrame):
    """
    Converts a properly formatted pandas dataframe with the columns
    of cell and time to a networkx network where nodes are locations
    and edges are formed between nodes with that share a time. 
    
    Args:
        df: a pandas dataframe withe two columns cell and time

    Returns:
        A networkx network
    """
    import networkx as nx
    net = nx.Graph()

    grouped = df.groupby('time')

    for name, group in grouped:
        cells = group['cell'].tolist()
        for i, cell_i in enumerate(cells):
            for cell_j in cells[i+1:]:
                if net.has_edge(cell_i, cell_j):
                    # if the edge already exists, increment the weight by 1
                    net[cell_i][cell_j]['weight'] += 1
                else:
                    # otherwise, add the edge with a weight of 1
                    net.add_edge(cell_i, cell_j, weight=1)

    return net


def convert_geosynchnet(fcol: FeatureCollection, time_delta: timedelta,
     hash_func: Callable):
    """
    Converts a FeatureCollection into a chronnet with a specified time interval
    using a specified hashing function
    
    Args:
        fcol: a FeatureCollection with time bound shapes 

        time_delta: the desired time interval

        hash_func: the hashing function

    Returns:
        A networkx network 
    """
    track = Track(fcol.geoshapes)
    timestamps = get_timestamp_intervals(track, time_delta)

    track_list = time_slice_track(track, timestamps)

    df = hash_tracks_into_netdf(track_list, timestamps, hash_func)

    geosynchnet = geosynchnet_create(df)

    return geosynchnet



""" Geotime hash representation"""
from typing import Callable, List
from collections import Counter
from datetime import  datetime, timedelta
from geostructures.typing import GeoShape
from geostructures.time import TimeInterval
from geostructures import FeatureCollection, Track

def precision_delta(precision: int):
    """
    Translates a timehash precision into the corresponding timedelta. 
    Args:
        precision: the desired precision of the timehash precision higher
        then 10 is not supported

    Returns:
        A timedelta
    """
    if precision == 1:
        delta = timedelta(days = 5840)
    elif precision == 2:
        delta = timedelta(days = 730)
    elif precision == 3:
        delta = timedelta(days = 91.2)
    elif precision == 4:
        delta = timedelta(days = 11.4)
    elif precision == 5:
        delta = timedelta(hours = 34.2)
    elif precision == 6:
        delta = timedelta(hours = 4.2)
    elif precision == 7:
        delta = timedelta(minutes = 32)
    elif precision == 8:
        delta = timedelta(minutes = 4)
    elif precision == 9:
        delta = timedelta(seconds = 30)
    elif precision == 10:
        delta = timedelta(seconds = 3.6)
    else:
        print("Only precision up to 10 is supported")
    return delta

def generate_times(start_time: datetime, end_time: datetime, precision: int):
    """
    Get the corresponding time list between two times with the
    interval determined by the timehash precision. 
    
    Args:
        start_time: the start time
        
        end_time: the endtime

        precision: the desired precision of the timehash precision higher
        then 10 is not supported

    Returns:
        A list of times
    """
    start = start_time
    end = end_time
    time_list = []

    while start <= end:
        time_list.append(start.timestamp())
        start += precision_delta(precision)

    return time_list




def timehash_geoshape(geoshape: GeoShape, precision: int):
    """
    Converts a geoshape time into a  list of timehashes of varying precision. 
    A single timepoint produces a list of 1.  
    Time intervals produces a list of geotime hashes that reflect the timehashes
    that cover that time interval at the given precision.
    Args:
        geoshape: a geoshape

        precision: the desired precision of the timehash precision higher
        then 10 is not supported

    Returns:
        A timehash list
    """
    # pylint: disable=import-outside-toplevel
    import timehash
    timehash_list = []
    assert isinstance(geoshape.dt, TimeInterval)
    start_time = geoshape.dt.start
    end_time = geoshape.dt.end

    time_list = generate_times(start_time, end_time, precision)

    for time in time_list:
        timehash1 = timehash.encode(time, precision)
        timehash_list.append(timehash1)

    return timehash_list


def append_timehash_to_geohashmap(hashmap: dict, timehash_list: List):
    """
    Appends a timehash to each geohash. If there are multiple timehashes in the list
    every permutation of geohash/timehash is appended.
    Args:
        hashmap: the hashmap produced by a geographic hashing function 

        timehash_list: the associated timehashes 

    Returns:
        A geotime hashmap
    """
    geotime_hashmap = {}
    for key, value in hashmap.items():
        for item in timehash_list:
            new_key = str(key) + "_" + str(item)
            geotime_hashmap[new_key] = value
    return geotime_hashmap


def breakdown_hashmap_by_suffix(hashmap: Counter):
    """
    Breaks down a hashmap by the common suffixes. 
    Args:
        hashmap: the target hashmap 

    Returns:
        A dictionary of dictionaries with each shared suffix serving
        as the outer key
    """
    suffix_dict = {}
    for key in hashmap:
        suffix = key.split('_')[-1]  # assuming suffixes are defined after the last underscore
        if suffix not in suffix_dict:
            suffix_dict[suffix] = {key: hashmap[key]}
        else:
            suffix_dict[suffix][key] = hashmap[key]

    return suffix_dict


def combine_dicts(d:dict):
    """
    Combines a dictionary of dictionaries by dropping the outer keys
    and flattening the inner dictionaries.
    Args:
        d: A dictionary of dictionaries 

    Returns:
        A dictionary of dictionaries with each shared suffix serving
        as the outer key
    """
    combined_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            combined_dict.update(v)
        else:
            combined_dict[k] = v
    return combined_dict


def convert_geotimehash(fcol: FeatureCollection, precision: int,
     hash_func: Callable):
    """
    Converts a FeatureCollection into a chronnet by a specified timehash precision
    using a specified hashing function
    
    Args:
        fcol: a FeatureCollection with time bound shapes 

        precision: the precision of the time hash

        hash_func: the hashing function


    Returns:
        A geotime hashmap
    """
    track = Track(fcol.geoshapes)
    master_geotime_hashmap:Counter = Counter()
    shape_count:Counter = Counter()

    for shape in track:
        geohashmap = hash_func(Track([shape]))
        timehash_list = timehash_geoshape(shape, precision)
        timehash_dict = {item: 1 for item in timehash_list}
        temp_geotime_hashmap = append_timehash_to_geohashmap(geohashmap, timehash_list)
        shape_count.update(timehash_dict)
        master_geotime_hashmap.update(temp_geotime_hashmap)

    suffix_dict = breakdown_hashmap_by_suffix(master_geotime_hashmap)

    for outer_key, inner_dict in suffix_dict.items():
        num_shapes = shape_count[outer_key]
        for inner_key, value in inner_dict.items():
            inner_dict[inner_key] /= num_shapes

    return combine_dicts(suffix_dict)



import datetime as dt
import pandas as pd
from geochron.geosynchnet import geosynchnet_create, convert_geosynchnet
from geostructures import Coordinate, GeoPoint
from geostructures.collections import  Track
from geostructures.geohash import H3Hasher

hasher = H3Hasher(resolution = 10)

def test_geosynchnet_create():
    test_dict={'cell': {0: '8a194ad32167fff',
    1: '8a194ad32b07fff',
    2: '8a194ad3056ffff',
    3: '8a194ad3078ffff',
    4: '8a194ad3078ffff'},
    'time': {0: '2020-01-01 08:05:00, 2020-01-01 09:05:00',
    1: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    2: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    3: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    4: '2020-01-01 10:05:01, 2020-01-01 11:05:01',}}

    test_df = pd.DataFrame(test_dict)
    test_network = geosynchnet_create(test_df)

    assert list(test_network)[0] == '8a194ad32b07fff'
    assert test_network.has_edge('8a194ad32b07fff', '8a194ad3056ffff') == True
    assert test_network.has_edge('8a194ad32167fff', '8a194ad3078ffff') == False
    assert test_network['8a194ad32b07fff']['8a194ad3056ffff']['weight'] == 1

def test_convert_geosynchnet():
    track = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5)),
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
    )

    test_geosynchnet = convert_geosynchnet(track, dt.timedelta(hours=1), hasher.hash_collection)
    
    assert list(test_geosynchnet)[0] == '8a194ad32b07fff'
    assert test_geosynchnet.has_edge('8a194ad32b07fff', '8a194ad3056ffff') == True


'Pydeck helpers'
import h3
from networkx import Graph

def h3edge_to_coordinates(edge:tuple):
    """
    Creates a tuple with coordinate locations from a tuple containing h3
    hexes and weight representing a geographic network edge.

    Args:
        edge: A tuple in the form (h3encoding_origin:str, 
        h3encoding_destination:str, edge_weight:int) 

    Returns:
        A tuple in the form of (coordinate_origin:float, 
        coordinate_destination:float, edge_weight:int)
    """

    edge_coordinate = (h3.h3_to_geo(edge[0]), h3.h3_to_geo(edge[1]), edge[2])

    return edge_coordinate


def convert_to_pydeckdict(edgecoordinates:tuple, origin:str, destination:str):
    """
    Converts a tuple with coordinate locations from a tuple containing h3
    hexes and weight representing a geographic network edge.

    Args:
        edgecoordinates: A tuple in the form of 
        (coordinate_origin:float, coordinate_destination:float, edge_weight:int)
        origin: the origin hex in string format
        destination: the origin hex in string format

    Returns:
        A dictionary for pydeck consumption
    """
    return {
        "start_lat": edgecoordinates[0][0],
        "start_lng": edgecoordinates[0][1],
        "end_lat": edgecoordinates[1][0],
        "end_lng": edgecoordinates[1][1],
        "width": edgecoordinates[2],
        "origin": origin,
        "destination": destination
    }



def network_arc_circle(network:Graph):
    """
    Creates data suitable for pydeck visualization in the Great Circle and Arc layers
    from a chron net or geosyync network that uses h3 hashing.

    Args:
        network: A networkx network

    Returns:
        A list of dictionaries for pydeck consumption
    """
    pydeck_data = []
    edgelist = list(network.edges.data('weight'))
    for edge in edgelist:
        edgecoordinates = h3edge_to_coordinates(edge)
        dict_edge = convert_to_pydeckdict(edgecoordinates, edge[0], edge[1])
        pydeck_data.append(dict_edge)
    return pydeck_data



import datetime as dt
import networkx as nx
import pandas as pd
from geochron.chronnet import hash_tracks_into_netdf, chronnet_create, convert_chronnet
from geostructures import Coordinate, GeoPoint
from geostructures.collections import  Track
from geostructures.geohash import H3Hasher

hasher = H3Hasher(resolution = 10)
test_timestamps = [dt.datetime(2020, 1, 1, 9, 5, tzinfo=dt.timezone.utc),
dt.datetime(2020, 1, 1, 10, 5, 1, tzinfo=dt.timezone.utc)]

def test_hash_tracks_into_netdf():

    track1 = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5))
    ]
    )
    track2 = Track(
    [
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
    )

    track_list = [track1, track2]
    test_df = hash_tracks_into_netdf(track_list, test_timestamps, hasher.hash_collection)

    assert test_df['cell'].values[4] == '8a194ad3078ffff'

def test_chronnet_create():
    test_dict={'cell': {0: '8a194ad32167fff',
    1: '8a194ad32b07fff',
    2: '8a194ad3056ffff',
    3: '8a194ad3078ffff',
    4: '8a194ad3078ffff'},
    'time': {0: '2020-01-01 08:05:00, 2020-01-01 09:05:00',
    1: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    2: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    3: '2020-01-01 09:05:00, 2020-01-01 10:05:01',
    4: '2020-01-01 10:05:01, 2020-01-01 11:05:01',}}

    test_df = pd.DataFrame(test_dict)

    test_network = chronnet_create(test_df, True, "directed")
    test_network_ud = chronnet_create(test_df, True, "undirected")
    no_self_loop = chronnet_create(test_df, False, "directed")

    assert list(test_network)[0] == '8a194ad3056ffff'
    assert test_network_ud.is_directed() == False
    assert list(nx.selfloop_edges(test_network )) == [('8a194ad3078ffff', '8a194ad3078ffff')]
    assert list(nx.selfloop_edges(no_self_loop)) == []

def test_convert_chronnet():
    track = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5)),
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
    )

    test_chronnet = convert_chronnet(track, dt.timedelta(hours=1), hasher.hash_collection, True, "directed")
    
    assert list(test_chronnet)[0] == '8a194ad3056ffff'
    



"""
Intercepts import errors concerning optional imports to provide a more detailed error response
"""

__all__ = ['ConditionalPackageInterceptor']

from importlib import util
import subprocess
import sys
from typing import Union


class ConditionalPackageInterceptor:  # pragma: no cover
    """
    Provides automatic pip installation of a package if it isn't found. Only packages added to this
    object using the .permit_packages() method are allowed to be automatically installed.

    To use:
        In your code's entrypoint, add the following code:

            ConditionalPackageInterceptor.permit_packages(
                <list or dict of packages>
            )
            sys.meta_path.append(ConditionalPackageInterceptor)

    Use Case:
        If you're building a library where only some of the functionality relies on third-party
        packages, and you anticipate that only some users will ever use that functionality, then
        you may want to only install those packages if they're actually needed.

    Note that this object needs to be defined before all of your other code runs or else it won't
    work. In most cases, the closest-to-root __init__.py is a great place.

    """

    PERMITTED_PACKAGES: dict = {}
    AUTO_DOWNLOAD = False

    @classmethod
    def permit_packages(cls, packages: Union[list, dict]) -> None:
        """
        Adds python packages to the list of packages that are permitted to be automatically
        installed.

        The name of a python package is not always the same between "pip install <package>" and
        "import <package>". Blindly installing any package that isn't found can therefore be not
        just a programmatic risk but a security one as well (malicious packages do exist).

        You can add to this list in two ways:
            As a list: packages will be pip installed exactly as listed
                ["xlrd"]
                "import xlrd" -> pip install xlrd

            As a dict: packages will be pip installed by the corresponding key
                {"yaml": "pyyaml"}
                "import yaml" -> pip install pyyaml

        Args:
            packages (Union[list, dict]): The packages that will be allowed to auto-install if
                                          missing

        Returns:
            None
        """
        if isinstance(packages, list):
            cls.PERMITTED_PACKAGES.update({item: item for item in packages})
        elif isinstance(packages, dict):
            cls.PERMITTED_PACKAGES.update(packages)
        else:
            raise TypeError(
                f"Permitted packages must be submitted as a list or dict, not {type(packages)}"
            )

    @classmethod
    def permit_auto_download(cls, option: bool) -> None:
        """
        Defines whether packages may be auto-downloaded or not. Default False.
        """
        cls.AUTO_DOWNLOAD = option

    @classmethod
    def find_spec(  # pylint: disable=unused-argument, inconsistent-return-statements
            cls, name, path, target=None
    ):
        """
        DO NOT USE.

        Will be executed by python's importlib when it's search for installed packages. This object
        will be examined last since you .append()'ed it to the sys.meta_path (you did follow the
        instructions in the docstring, right?), which means that if importlib got this far without
        finding the package then the package doesn't exist in your virtual environment.

        Will only pip install if the package name is in the PERMITTED_PACKAGES class variable.
        Otherwise you'll get the usual "ModuleNotFoundError".

        Args:
            name (str): The name of the package
            path:
            target:

        Returns:

        """
        if name not in cls.PERMITTED_PACKAGES:
            return

        if cls.AUTO_DOWNLOAD:
            print(f"Module {name!r} not installed. Attempting to pip install...")
            try:
                subprocess.run(
                    f"{sys.executable} -m pip install {cls.PERMITTED_PACKAGES[name]}",
                    check=True
                )
            except subprocess.CalledProcessError:
                return None

            return util.find_spec(name)

        raise ModuleNotFoundError(
            f"You are attempting to use a module which requires an optional installation ({name}). "
            "Please execute the following command to continue: \n "
            f"    pip install {cls.PERMITTED_PACKAGES[name]}"
        )

